-module(pov_exp).
-import(lists, [foldl/3,map/2,foreach/2,reverse/1,seq/2,
                flat_length/1,append/1,append/2,keydelete/3]).
-import(kayos_utils, [db/1,db/2,foreach_except_last/3,foreach_except_last_incr/4,comma_separated_list/3]).
-import(wpc_pov, [get_var/1]).
-export([export/3]).
-include("e3d.hrl").
-include("e3d_image.hrl").
-include("pov.hrl").



%% conditonal_export was created for those hand editing their POVRay files
%% It is typically used when writing sections or objects to different files
%% To read the correct file without modifying it. This would preserve say a
%% hand modified lighting setup while allowing the user to change the
%% Material qualities -- Please Rewrite
%% Geometry & Material sections is handled slightly differently.

conditional_export(F,Section,ExportFunction,Base,Root,Attr) ->
  SplitFiles = proplists:get_value(split,Attr,single),
  SectionStr = atom_to_list(Section),
  io:put_chars("Section: " ++ SectionStr ++"\n"),
  UsePredefined = list_to_atom("predefinded_" ++ SectionStr),
  PredefinedFile = list_to_atom(SectionStr ++ "_file"),
  case proplists:get_value(UsePredefined,Attr,false) of
    true ->
      io:format(F, "#include ~p\n", [proplists:get_value(PredefinedFile,Attr,"")]);
    false ->
      case proplists:get_value(Section,Attr,true) of
        true ->
          case SplitFiles of
            single ->
              ExportFunction(F,Attr);
            _      ->
              ExportRoot = Root ++ "_" ++ SectionStr ++ ".inc",
              ExportName = Base ++ "_" ++ SectionStr ++ ".inc",
              {ok,ExportFile} = file:open(ExportRoot, [write]),
              unofficial_version(ExportFile,Attr),
              ExportFunction(ExportFile,Attr),
              io:format(F, "Include ~p\n", [ExportName]),
              ok = file:close(ExportFile),
              io:put_chars("Wrote " ++ SectionStr ++ " section\n")
            end;
        false ->
          case SplitFiles of
            single -> ok;
            _      ->
              ExportName = Base ++ "_" ++ atom_to_list(Section) ++ ".pov",
              io:format(F, "#include ~p\n", [ExportName])
            end,
          io:put_chars("Section " ++ SectionStr ++ " not needed\n")
        end
      end.


unofficial_version(F, Attr) ->
  case proplists:get_value(version, Attr) of
    megapov -> io:put_chars(F, "#version unofficial MegaPov 0.7;\n");
    mlpov   -> io:put_chars(F, "#version unofficial mlpov 0.8;\n");
    _       -> true end.

use_sss(F, Attr)  ->
  case proplists:get_value(use_sss, Attr) of
    true -> io:put_chars(F, "#include \"sss.inc\"\n");
    false -> ok end.

%-------------------------------------------------------------------%
export(Name, #e3d_file{objs=Objs,mat=Mat,creator=Creator}, Attr	) ->
%-------------------------------------------------------------------%
  TmpDir = proplists:get_value(pov_tmpdir,Attr,none),
  case TmpDir of none -> ok; _ -> file:make_dir(TmpDir) end,
  {ok,F} = file:open(Name, [write]),
  Root = filename:rootname(Name, ".pov"),
  Base = filename:basename(Root),
  unofficial_version(F, Attr),
  io:format(F, "// Exported from ~s\n", [Creator]),
  use_sss(F, Attr),

%--------------------- World Stuff  --------------------------------%
  io:format("split: ~p~n",[get_var(split)]),
  ExportWorld = fun(F,Attr) -> export_world(F,Attr) end,
  conditional_export(F,world,ExportWorld,Base,Root,Attr),

%--------------------- Camera Stuff --------------------------------%
  ExportCamera  = fun(F,Attr) -> export_camera(F,Attr) end,
  conditional_export(F,camera,ExportCamera,Base,Root,Attr),

%----------------------- Lighting Stuff ----------------------------%
  ExportLights = fun(F,Attr) -> export_lights(F,Attr) end,
  conditional_export(F,lighting,ExportLights,Base,Root,Attr),

%----------------------- Materials Stuff ---------------------------%
%fprof:trace(start),
  case proplists:get_value(materials, Attr, true) of
    true ->
      case proplists:get_value(split, Attr, single) of
        single ->
          export_materials(F, Mat, Base, Root, Attr);
        _->
          MaterialsName = Base ++ "_materials.inc",
          MaterialsRoot = Root ++ "_materials.inc",
          {ok, MaterialsFile} = file:open(MaterialsRoot, [write]),
          unofficial_version(MaterialsFile,Attr),
          export_materials(MaterialsFile, Mat, Base, Root, Attr),
          io:format(F, "#include ~p\n", [MaterialsName]),
          ok = file:close(MaterialsFile)
        end,
      io:put_chars("Wrote Materials\n");
    false ->
      case proplists:get_value(split, Attr, single) of
        single ->
          ok;
        _ ->
          MaterialsName = Base ++ "_materials.inc",
          io:format(F, "#include ~p\n", [MaterialsName])
          end,
        io:put_chars("Materials not needed\n")
        end,
 %fprof:trace(stop), 
 %fprof:profile(),
 %fprof:analyse([{dest, "wpc_pov.analysis"}]),
  
%----------------------- Geometry Stuff ----------------------------% 

  case proplists:get_value (split, Attr, single) of
    single -> 
      foreach(fun(Obj) ->
        #e3d_object{name=ObjName0} = Obj,
        ObjName = kayos_utils:safe_name(ObjName0),
        export_geometry(F, Obj, Mat, Base, Attr),
        io:put_chars(F,ObjName),
        io:put_chars(F,"\n")
      end, Objs);
    bysection ->
      GeometryName = Base ++ "_geometry.inc",
      GeometryRoot = Root ++ "_geometry.inc",
      {ok,GeometryFile} = file:open(GeometryRoot, [write]),
      unofficial_version(GeometryFile,Attr),
      io:format(F, "#include ~p\n", [GeometryName]),
      foreach(fun(Obj) ->
        #e3d_object{name=ObjName0} = Obj,
        ObjName = kayos_utils:safe_name(ObjName0),
        export_geometry(GeometryFile, Obj, Mat, Base, Attr),
        io:put_chars(F,ObjName),
        io:put_chars(F,"\n")
      end, Objs),
      ok = file:close(GeometryFile);
    byobject ->
      foreach(fun(Obj) ->
        #e3d_object{name=ObjName0} = Obj,
        ObjName = kayos_utils:safe_name(ObjName0),
        GeometryName = Base ++ "_" ++ ObjName ++ ".inc",
        GeometryRoot = Root ++ "_" ++ ObjName ++ ".inc",
        {ok,GeometryFile} = file:open(GeometryRoot, [write]),
        unofficial_version(GeometryFile,Attr),
        io:format(F, "#include ~p\n", [GeometryName]),
        export_geometry(GeometryFile, Obj, Mat, Base, Attr),
        io:put_chars(F,ObjName),
        io:put_chars(F,"\n"),
      ok = file:close(GeometryFile)
      end, Objs)
  end,
  io:put_chars("Wrote Geometry\n\n"),
  ok = file:close(F).
  
%-------------------------------------------------------------------%

export_world(F,Attr) ->
  case proplists:get_value(radiosity,Attr,false) of
    true -> io:put_chars(F, "#include \"rad_def.inc\"\n"); _ -> ok end,
  io:put_chars(F, "\nglobal_settings {\n"),

  ambient_lighting(F,proplists:get_value(lights, Attr)),
  case proplists:get_value(radiosity,Attr,false) of
    true ->
      io:put_chars(F, "  radiosity {\n"),
      Preset = case proplists:get_value(gi_preset,Attr,default) of
        default -> "Radiosity_Default";
        debug -> "Radiosity_Debug";
        fast -> "Radiosity_Fast";
        normal -> "Radiosity_Normal";
        twobounce -> "Radiosity_2Bounce";
        final -> "Radiosity_Final";
        outdoorlq -> "Radiosity_OutdoorLQ";
        outdoorhq -> "Radiosity_OutdoorHQ";
        outdoorlight -> "Radiosity_OutdoorLight";
        indoorlq -> "Radiosity_IndoorLQ";
        indoorhq -> "Radiosity_IndoorHQ"
      end,
      io:put_chars(F, "      Rad_Settings (" ++ Preset ++ ", off, off)\n"),
      io:put_chars(F, "  }\n");
    false -> ok end,
  io:put_chars(F, "}\n"),
  case proplists:get_value(background,Attr,false)of
    true ->
      {R,G,B}= proplists:get_value(bg_color,Attr,{0.0,0.0,0.0}),
      io:format(F, "  background {<~p, ~p, ~p>}\n",[R,G,B]);
    false ->
      ok end.
  
ambient_lighting(F, [{_,Ps}|Ls]) ->
  io:put_chars("hi\n"),
  OpenGL = proplists:get_value(opengl, Ps, []),
  Type = proplists:get_value(type, OpenGL, point),
  io:format("Type ~p\n",[Type]),
  case Type of
    ambient ->
       {R,G,B,_} = proplists:get_value(diffuse, OpenGL, {0.0,0.0,0.0,0.0}),
       io:format(F, "  ambient_light <~f, ~f, ~f>\n", [R,G,B]);
    _ -> 
      ambient_lighting(F, Ls)
  end;
ambient_lighting(_F, []) -> ok.
      
export_camera(F, Attr) ->
  #camera_info{aim=Aim,distance_to_aim=Distance,elevation=El,azimuth=Az,tracking={TrackX,TrackY},fov=Fov} =
    proplists:lookup(camera_info, Attr),
  Width = proplists:get_value(frame_width,Attr,320),
  Height = proplists:get_value(frame_height,Attr,240),

  Ro = math:pi()/180.0,
  Dist = limit_dist(Distance),
  Dy = Dist * math:sin(El*Ro),
  P = Dist * math:cos(El*Ro),
  Dx = -(P * math:sin(Az*Ro)),
  Dz = P * math:cos(Az*Ro),
  Rev = {Dx,Dy,Dz},
  %L = e3d_vec:norm_cross(Rev, {0.0,1.0,0.0}),
  L=e3d_vec:norm(-Dz,0.0,Dx),
  LeftN = case e3d_vec:is_zero(L) of
    false ->
      L;
    true ->
      { -math:cos(Az*Ro), 0.0, -math:sin(Az*Ro)}
  end,
  {Lx,Ly,Lz} = LeftN,
  DownN = e3d_vec:norm(Dy*Lz-Dz*Ly, Dz*Lx-Dx*Lz, Dx*Ly-Dy*Lx),
  Transl = e3d_vec:add(
    e3d_vec:mul(LeftN, TrackX),
    e3d_vec:mul(DownN, TrackY)),
  Pos = e3d_vec:sub(Rev,Aim),
  From = e3d_vec:add(Pos, Transl),
  To = e3d_vec:sub(Transl, Aim),
  %Up = e3d_vec:sub(From, DownN),

  io:put_chars(F, "camera {\n"),
  case proplists:get_value(ortho_cam,Attr,false) of
    true -> io:put_chars(F, "  orthographic\n");
    _ -> io:put_chars(F, "  perspective\n")
  end,
  export_pos(F, location, From),
  io:format(F, "  right < -~p, 0, 0>~n", [Width/Height]),
  %export_pos(F, up, Up),
  export_pos(F, look_at, To),
  io:format(F, "  angle ~p~n",[Fov]),
  io:put_chars(F, "}\n").

limit_dist(Dist) when Dist > 0.0 ->
    if Dist < 0.01 -> 0.01;
       true -> Dist
    end;
limit_dist(Dist) when Dist < 0.0 ->
    if Dist > -0.01 -> -0.01;
       true -> Dist
    end;
limit_dist(_) -> 0.01.

export_pos(F,Str,{X,Y,Z}) ->
  io:format(F, "  ~s <~.10f, ~.10f, ~.10f>~n", [Str,X,Y,Z]).

%----- LIGHTS ----------------%

export_lights(F,Attr) ->
  Ls = proplists:get_value(lights, Attr),
  foreach(fun(L) -> export_light(F, L) end, Ls).

export_light(F, {_,Ps}) ->
  OpenGL = proplists:get_value(opengl, Ps, []),
  Povray = proplists:get_value(povray, Ps, []),
  Type0 = proplists:get_value(type, OpenGL, point),

  case Type0 of
   ambient -> ok;
    _ ->
      Type = case proplists:get_value(cylinder,Povray, false) andalso Type0 == spot of
        true -> cylinder;
        false -> Type0 end,
      export_light_common(F, OpenGL),
      case proplists:get_value(shadowless, Povray, false) of
        true -> io:put_chars(F, "  shadowless\n");
        false -> ok end,
      case proplists:get_value(area_light, Povray, false) of
        true -> export_light(F, area, OpenGL,Povray); false -> ok end,
      Type = case proplists:get_value(cylinder,Povray, false) andalso Type0 == spot of
        true -> cylinder;
        false -> Type0 end,
      export_light(F, Type, OpenGL, Povray)
   end.

export_light(F, area, OpenGL, Povray) ->
  io:put_chars(F,"  area_light\n"),
  From = proplists:get_value(from, OpenGL, {0.0,0.0,1.0}),
  To = proplists:get_value(from, OpenGL, {0.0,0.0,0.0}),
  L1 = proplists:get_value(area_height, Povray, 2.0)/2,
  L2 = proplists:get_value(area_width, Povray, 2.0)/2,
  {V1x,V1y,V1z,V2x,V2y,V2z} = area_light_vectors(From,To,L1,L2),
  io:format(F, "  <~p, ~p, ~p>, <~p, ~p, ~p>, ~p, ~p\n", [V1x,V1y,V1z,V2x,V2y,V2z,
                                                          proplists:get_value(area_lights_x,Povray,2),
                                                          proplists:get_value(area_lights_y,Povray,2)]),
  io:format(F, "  adaptive ~p\n", [proplists:get_value(area_adaptive, Povray, 1)]),
  case proplists:get_value(area_jitter, Povray, false) of true -> io:put_chars(F, " jitter\n"); false -> ok end,
  case proplists:get_value(area_circular, Povray, false) of true -> io:put_chars(F, " circular\n"); false -> ok end,
  case proplists:get_value(area_orient, Povray, false) of true -> io:put_chars(F, " orient\n"); false -> ok end;
export_light(F, point, _OpenGL, Povray) ->
  export_light_options(F,Povray),
  io:put_chars(F, "\n}\n");
export_light(F, infinite, OpenGL, Povray) ->
  To = proplists:get_value(aim_point, OpenGL, {0,0,1}),
  io:put_chars(F, "\n  parallel\n"),
  show_point(F, "  point_at", To),
  export_light_options(F,Povray),
  io:put_chars(F, "\n}\n");
export_light(F, spot, OpenGL, Povray) ->
  To = proplists:get_value(aim_point, OpenGL, {0,0,1}),
  Angle = proplists:get_value(cone_angle, OpenGL, 30),
  Falloff = proplists:get_value(spot_exponent, OpenGL, 0),
  io:format(F, "\n  spotlight \n  radius ~p \n  falloff ~p \n  tightness ~p \n", [Angle,Angle+Falloff,proplists:get_value(tightness, Povray, 0.0)]),
  show_point(F, "  point_at ", To),
  export_light_options(F,Povray),
  io:put_chars(F, "\n}\n");
export_light(F, cylinder, OpenGL, Povray) ->
  To = proplists:get_value(aim_point, OpenGL, {0,0,1}),
  Angle = proplists:get_value(cone_angle, OpenGL, 30),
  Falloff = proplists:get_value(spot_exponent, OpenGL, 0),
  io:format(F, "\n  cylinder \n  radius ~p \n  falloff ~p \n  tightness ~p \n", [Angle,Angle+Falloff,proplists:get_value(tightness, Povray, 0.0)]),
  show_point(F, "  point_at", To),
  export_light_options(F,Povray),
  io:put_chars(F, "\n}\n");
export_light(F, Type, _, Povray) ->
  export_light_options(F,Povray),
  io:format("unknown light type defaulting to point: ~p\n", [Type]),
  io:put_chars(F, "\n}\n").

export_light_common(F, OpenGL) ->
  From = proplists:get_value(position, OpenGL, {0,0,0}),
  {R,G,B,_} = proplists:get_value(diffuse, OpenGL, {1,1,1,1}),
  io:put_chars(F, "light_source { \n  "),
  show_point(F,From), io:put_chars(F, "\n  "),
  show_point(F,"rgb",{R,G,B}).

export_light_options(F, Povray) ->
  case proplists:get_value(fade_distance, Povray, false) of
    true ->
      io:format(F, "\n  fade_distance ~p", [proplists:get_value(fade_distance_amt, Povray, 0.5)]),
      io:format(F, "\n  fade_power ~p", [proplists:get_value(pov_falloff, Povray, 0)]);
    false -> ok end,
  case proplists:get_value(atmo_interaction, Povray, true) of
    true -> io:put_chars(F, "\n  media_interaction "); false -> ok end,
  case proplists:get_value(atmo_att, Povray, false) of
    true -> io:put_chars(F, "\n  media_attenuation "); false -> ok end.

area_light_vectors({X2,Y2,Z2},{X1,Y1,Z1}, L1, L2) ->
  Az = get_angle({0.0,0.0,1.0},{X2-X1,0.0,Z2-Z1}),
  El = get_angle({X2-X1,0.0,Z2-Z1},{X2-X1,Y2-Y1,Z2-Z1}),

  X3 = X2 - (L1 * math:sin(El) * math:sin(Az)),
  Y3 = Y2 + (L1 * math:cos(El)),
  Z3 = Z2 - (L1 * math:sin(El) * math:cos(Az)),

  X4 = X2 + (L2 * math:cos(Az)),
  Y4 = Y2,
  Z4 = Z2 + (L2 * math:sin(Az)),

  X5 = X3 + X4 - X2,
  Y5 = Y3 + Y4 - Y2,
  Z5 = Z3 + Z4 - Z2,

  V1x = 2 * (X3 - X5),
  V1y = 2 * (Y3 - Y5),
  V1z = 2 * (Z3 - Z5),

  V2x = 2 * (X4 - X5),
  V2y = 2 * (Y4 - Y5),
  V2z = 2 * (Z4 - Z5),

  {V1x, V1y, V1z, V2x, V2y, V2z}.

get_angle({X2,Y2,Z2},{X1,Y1,Z1}) ->
  A_dot_B = X2*X1 + Y2*Y1 + Z2*Z1,
  %io:format("A.B: ~p\n", [A_dot_B]),
  Ln_A = math:sqrt(X1*X1 + Y1*Y1 + Z1*Z1),
  Ln_B = math:sqrt(X2*X2 + Y2*Y2 + Z2*Z2),
  io:format("Ln_A: ~p\n", [Ln_A]),
  io:format("Ln_B: ~p\n", [Ln_B]),
  Ln_AB = Ln_A*Ln_B,
  Angle = case Ln_AB == 0 of
    true -> 0.0;
    false -> A_dot_B/Ln_AB end,
  io:format("Angle: ~p\n", [Angle]),
  io:put_chars("this side\n"),
  math:acos(Angle).

show_point(F, Label, {X,Y,Z}) ->
  io:put_chars(F, Label),
  io:format(F, " <~p, ~p, ~p> ", [X,Y,Z]).
show_point(F,{X,Y,Z}) ->
  io:format(F, "<~p, ~p ~p> ", [X,Y,Z]).

%---------------- MATERIALS --------------------------------%

export_materials(F, [{Name0,Mat}|T], Base, Root, Attr) ->
  db("Root2",Root),
  Name = kayos_utils:safe_name(atom_to_list(Name0)),
  Maps = proplists:get_value(maps, Mat),
  Tmpdir = proplists:get_value(pov_tmpdir, Attr, none),
  db("tmpdir2",Tmpdir),
  case Tmpdir of
    none ->
      export_maps(F, Maps, Root, Name),
      export_material(F, Name, Mat, Base, Attr),
      export_materials(F, T, Base, Root, Attr);
    _ ->
      export_maps(F, Maps, Tmpdir, Name),
      export_material(F, Name, Mat, Tmpdir, Attr),
      export_materials(F, T, Base, Root, Attr)
  end;
export_materials(_F,[], _Base, _Root, _Attr) ->
  ok.

export_maps(F, [{diffuse,Map}|_T], Base, Name) ->
  db("Base3",Base),
  export_map(F, "Kd", Map, Base, Name);
export_maps(F, [{ambient,Map}|_T], Base, Name) ->
  export_map(F, "Ka", Map, Base, Name);
export_maps(F, [{bump,Map}|T], Base, Name) ->
  export_map(F, "Bump", Map, Base, Name),
  export_maps(F, T, Base, Name);
export_maps(F, [_|T], Base, Name) ->
  export_maps(F, T, Base, Name);
export_maps(_, [], _, _) -> ok.

export_map(_, _, none, _, _) -> none;
export_map(_F, Label0, #e3d_image{width=W,height=H,image=Map}=Image, Base, Name) ->
  Label = "map_" ++ Label0,
  MapFile = Base ++ "_" ++ Name ++ "_" ++ Label ++ ".tga",
%  Image = #e3d_image{image=Map,width=W,height=H},
  ok = e3d_image:save(Image, MapFile).

export_material(F, Name, Mat, Base, _Attr) ->
  io:put_chars("export_material\n"),
  OpenGL = proplists:get_value(opengl, Mat),
  Povray = proplists:get_value(povray, Mat, []),
  Maps = proplists:get_value(maps, Mat),
  {Dr,Dg,Db,Opacity0} = proplists:get_value(diffuse, OpenGL),
  Opacity = 1 - Opacity0,
  {Ar,Ag,Ab,_} = proplists:get_value(ambient, OpenGL),
  {Sr,Sg,Sb,_} = proplists:get_value(specular, OpenGL),
  Shine = proplists:get_value(shininess, OpenGL), 
  io:put_chars(F, "#declare " ++ Name ++ "_WingsMat =\n  texture{\n" ),
  case length(Maps) of
    0 ->
      io:format(F, "      pigment{ rgbf <~p, ~p, ~p, ~p>}\n", [Dr,Dg,Db,Opacity]);
    _ ->
      %% Fixme: May break when map types other than Diffuse are implemented.
      MapName = Base ++ "_" ++ Name ++ "_map_Kd.tga",
      io:format(F, "      pigment{ uv_mapping image_map{ tga ~p }}\n", [MapName])
  end, 
  
  case proplists:get_value(pov_use_bump_map, Povray, false) of
   true ->
     io:put_chars(F, "      normal{\n"),
     io:format(F, "   uv_mapping bump_map{ tga ~p }}\n", [proplists:get_value(pov_bump_map, Povray, "filename")]);
  false -> ok end,
     
  io:put_chars(F, "      finish{\n"),
  io:format(F, "        ambient rgb <~p, ~p, ~p>\n", [Ar,Ag,Ab]),
  io:format(F, "        diffuse ~p\n", [proplists:get_value(diffusion,Povray,0.6)]),
  io:format(F, "        brilliance ~p\n", [proplists:get_value(brilliance,Povray,1.0)]),
  case proplists:get_value(phong_hl,Povray,false) of
    true -> io:format(F, "        phong ~p phong_size ~p\n", 
      [proplists:get_value(phongamt,Povray,0.5),proplists:get_value(phongsize,Povray,40.0)]);   
    false -> ok end,
  case proplists:get_value(spec_hl,Povray,true) of
    true -> io:format(F, "        specular ~p roughness ~p\n",
      [proplists:get_value(specamt,Povray,0.5),proplists:get_value(roughness,Povray,0.1)]);
    false -> ok end,
  case proplists:get_value(metalic,Povray,false) of
    true -> io:format(F, "        metallic ~p\n",
      [proplists:get_value(metalicamt,Povray,1.0)]);
    false -> ok end, 
  case proplists:get_value(reflection,Povray,false) of
    true ->
      Min = case proplists:get_value(min_reflection,Povray,false) of 
        true -> float_to_list(proplists:get_value(min_reflection_amt,Povray,0.01)) ++ "," ; 
        false -> " " end,
      io:put_chars(F, "        reflection { " ++ Min),
      io:format(F, " rgb <~p, ~p, ~p> ", [Sr,Sg,Sb]),
      case proplists:get_value(fresnel,Povray,false) of true -> io:put_chars(F, "fresnel "); false-> ok end,
      case proplists:get_value(falloff,Povray,false) of 
        true -> io:format(F, "falloff ~p ", [proplists:get_value(falloff_amt,Povray,0.5)]);
        false -> ok end,
      case proplists:get_value(exponent,Povray,false) of 
        true -> io:format(F, "exponent ~p ", [proplists:get_value(exponent_amt,Povray,0.5)]);
        false -> ok end,  
      case proplists:get_value(metalic_ref,Povray,false) of 
        true -> io:format(F, "metallic ~p ", [proplists:get_value(metalic_ref_amt,Povray,1.0)]);
        false -> ok end,
      io:put_chars(F, "}\n"),  
      case proplists:get_value(conserve_energy,Povray,false) of 
        true -> io:put_chars("        conserve_energy\n");
        false -> ok end; 
    false -> ok end, 
  case proplists:get_value(iridescence,Povray,false) of
    true -> 
      io:format(F, "        irid { ~p ", [proplists:get_value(iridescence_amt,Povray,0.12)]),
      case proplists:get_value(irid_thickness,Povray,false) of 
        true -> io:format(F, "thickness ~p ", [proplists:get_value(irid_thickness_amt ,Povray, 0.1)]);
        false -> ok end,
      case proplists:get_value(irid_turb,Povray,false) of 
        true -> io:format(F, "turbulence ~p ", [proplists:get_value(irid_turb_amt ,Povray, 0.1)]);
        false -> ok end,
      io:put_chars(F, "}\n");
    false -> ok end,
  io:put_chars(F, "        }\n}\n").
  
export_interior(F, Name, Povray) ->
  {FadeR, FadeG, FadeB} = proplists:get_value(fade_color,Povray,{0.0,0.0,0.0}),
  {AbsorpR, AbsorpG, AbsorpB} = proplists:get_value(interior_absorption,Povray,{0.0,0.0,0.0}),
  {EmitR, EmitG, EmitB} = proplists:get_value(interior_emission,Povray,{0.0,0.0,0.0}),
  {ScatterR, ScatterG, ScatterB} = proplists:get_value(interior_scattering,Povray,{0.0,0.0,0.0}),
  Hollow = proplists:get_value(hollow,Povray,false),
  ScatterType = proplists:get_value(scatter_type,Povray,1),
  
  %io:put_chars(F, "#declare " ++ Name ++ "_WingsInt =\n  interior{\n" ),

  case Hollow of 
  false ->
  io:put_chars(F, "  interior{\n" ),
    io:format(F, "    ior ~p\n", [proplists:get_value(pov_ior,Povray,1.0)]),
    io:format(F, "    dispersion ~p\n", [proplists:get_value(pov_dispersion_amt,Povray,0.0)]), 
    io:format(F, "    dispersion_samples ~p\n", [proplists:get_value(pov_dispersion_samples,Povray,7)]),
    io:format(F, "    fade_distance ~p\n", [proplists:get_value(fade_distance,Povray,0.5)]), 
    io:format(F, "    fade_power ~p\n", [proplists:get_value(fade_power,Povray,0)]),
    io:format(F, "    fade_color < ~p, ~p, ~p>\n", [FadeR, FadeG, FadeB]),
    io:format(F, "    caustics ~p\n", [proplists:get_value(pov_caustics,Povray,0.0)]);
  true ->
    io:put_chars(F, "    hollow\n"),
     io:put_chars(F, "  interior{\n" ),
    case proplists:get_value(interior_media,Povray,false) of
      true ->
        io:put_chars(F, "   media {\n"),
        io:format(F, "      absorption < ~p, ~p, ~p>\n", [AbsorpR, AbsorpG, AbsorpB]),
        io:format(F, "      emission < ~p, ~p, ~p>\n", [EmitR, EmitG, EmitB]),
        io:format(F, "      scattering { ~p, < ~p, ~p, ~p> extinction ~p ", [ScatterType, ScatterR, ScatterG, ScatterB, 
                                                                 proplists:get_value(extinction_amt,Povray,1.0)]),
        case ScatterType of 
          5 -> io:format(F, "eccentricity ~p }\n", [proplists:get_value(eccentricity_amt,Povray,0.0)]);
          _ -> io:put_chars(F, "}\n") end,
        io:put_chars(F, "    }\n"); 
      false -> 
        ok end
    end,
  io:put_chars(F, "}\n").


export_geometry(F, #e3d_object{name=Name0,obj=Mesh0}, Mat, Base, Attr) ->
  UseSSS = proplists:get_value(use_sss, Attr, false),
  db("UseSSS",use_sss),
  #e3d_mesh{tx=TxTmp} = Mesh0,
  Mesh1 = case proplists:get_value(optimise_uvs, Attr)
	andalso proplists:get_value(export_uv, Attr)
	andalso length(TxTmp) > 0 of
	  true -> kayos_utils:optimise_uvs(Mesh0);
	  _ -> Mesh0
	  end,
  Name = kayos_utils:safe_name(Name0),
  case UseSSS of
    true -> Mesh2 = e3d_mesh:triangulate(Mesh1); %quadrangulate(Mesh1); Swith back when quadrangulate becomes UV safe...
    false -> Mesh2 = e3d_mesh:triangulate(Mesh1) end,
  Mesh = e3d_mesh:vertex_normals(Mesh2),
  #e3d_mesh{fs=Fs0} = Mesh,
  Fs1 = [{M,FaceRec} || #e3d_face{mat=M}=FaceRec <- Fs0],
  Fs = sofs:to_external(sofs:relation_to_family(sofs:relation(Fs1))),
  #e3d_mesh{fs=Flist,vs=Vs,tx=Tx,he=He, ns=Ns} = Mesh,
  LnFs = length(Flist),
  ExportUVs = proplists:get_value(export_uv, Attr) andalso length(Tx) > 0,
  case proplists:get_value(gen_bounds, Attr) of
    true ->
	  {Xmin, Xmax, Ymin, Ymax, Zmin, Zmax} = kayos_utils:get_bounds(Vs),
	  io:format(F,"//  X goes from ~p to ~p - length: ~p\n", [Xmin, Xmax, Xmax - Xmin]),
	  io:format(F,"//  Y goes from ~p to ~p - length: ~p\n", [Ymin, Ymax, Ymax - Ymin]),
	  io:format(F,"//  Z goes from ~p to ~p - length: ~p\n", [Zmin, Zmax, Zmax - Zmin]);
	_ -> {Xmin, Xmax, Ymin, Ymax, Zmin, Zmax} = {0, 0, 0, 0, 0, 0}
	  end,
  Ml = material_list(Mat),
  io:format("ln Mat: ~p, ln Ml: ~p\n", [length(Mat), length(Ml)]),
  case UseSSS of
    true -> export_sss(F, Name, Vs, LnFs, Fs, ExportUVs, Ml, Mat, Tx, He, Attr);
    false -> export_mesh2(F, Name, Vs, LnFs, Fs, ExportUVs, Ml, Mat, Ns, Tx, Attr) end.

export_sss(F, Name, Vs, LnFs, Fs, ExportUVs, Ml, Mat, Tx, He, Attr)->
  db("Fs",Fs),
  io:format(F, "#declare ~s = object {\n", [Name]),
%---------------- Points Array ---------------%
  PointsArray = Name ++ "PA",
  start_array(F, PointsArray, length(Vs), 1),
  comma_separated_list( F, fun({X,Y,Z}) -> io:format(F, "<~p,~p,~p>", [X,Y,Z]) end, Vs),
  end_array(F),
%--------------- Faces Array -----------------%
  FacesArray = Name ++ "FA",
  start_array(F, FacesArray, LnFs, 4),
  comma_separated_list (F, fun({_Ms,Faces}) ->
    comma_separated_list (F, fun(#e3d_face{vs=Vs0}) ->
      case Vs0 of
        [A,B,C,D] -> io:format(F, "{~p,~p,~p,~p}", [A,B,C,D]);
        [A,B,C] -> io:format(F, "{~p,~p,~p,-1}", [A,B,C]) end end,
      Faces) end,
    Fs),
  end_array(F),
%---------------- UV Array ------------------%
  UVArray = Name ++ "UV",
  case ExportUVs of
    true ->
      Tx_tpl = list_to_tuple(Tx),
      start_array(F, UVArray, LnFs, 4),
      comma_separated_list(F, fun({_Ms,Faces}) ->
        db("Faces", Faces),
        comma_separated_list(F, fun(#e3d_face{tx=Tx0}) ->
	  db("Tx0",Tx0),
          case Tx0 of
	    [A,B,C,D] ->
	      {U0,V0} = element(A+1,Tx_tpl),{U1,V1} = element(B+1,Tx_tpl),
              {U2,V2} = element(C+1,Tx_tpl),{U3,V3} = element(D+1,Tx_tpl),
	      io:format(F, "{<~f,~f>,<~f,~f>,<~f,~f>,<~f,~f>}", [U0,V0,U1,V1,U2,V2,U3,V3]);
	    [A,B,C] ->
	      {U0,V0} = element(A+1,Tx_tpl),{U1,V1} = element(B+1,Tx_tpl),
	      {U2,V2} = element(C+1,Tx_tpl),
              io:format(F, "{<~f,~f>,<~f,~f>,<~f,~f>,<-1,-1>}", [U0,V0,U1,V1,U2,V2]) end end,
	  Faces) end,
	Fs),
      end_array(F);
    false -> ok end,
%-------------- Hard Edge Array -------------%
  HEArray = Name ++ "HE",
  LnHe = length(He),
  case LnHe of
    0 -> ok;
    _ ->
      start_array(F, HEArray, LnHe, 3),
      comma_separated_list(F, fun({A,B})-> io:format(F, "{~p,~p,-1}", [A,B]) end, He),
      end_array(F) end,
%-------------- Texture Index Array ---------------%
  TIArray = Name ++ "TI",
  start_array(F, TIArray, LnFs, 1),
  comma_separated_list (F, fun({_Ms,Faces}) ->
    comma_separated_list(F, fun(#e3d_face{mat=Mat0}) ->
      MatIndex = material_index(Mat0, Ml),
      io:format(F, "~p", [MatIndex]) end,
    Faces) end,
  Fs),
  end_array(F),
%-------------- Texture Array ---------------%
  TXArray = Name ++ "TX",
  start_array(F, TXArray, length(Mat), 1),
  comma_separated_list(F,
    fun({M0,Index}) ->
      SafeName = kayos_utils:safe_name(atom_to_list(M0)),
      TxDec = "  texture { " ++ SafeName ++ "_WingsMat }\n",
      io:put_chars(F, TxDec) end,
    Ml),
  end_array(F),
%-------------- Do It -----------------------%
  io:format(F, "  SSS_ImportBasicMesh(~s,~s) \n  SSS_AssignTextures(~s)\n", [PointsArray,FacesArray,TIArray]),
  case LnHe of
    0 -> ok; _ -> io:format(F, "  SSS_ImportSharpData(~s)\n", [HEArray]) end,
  case ExportUVs of
    true -> io:format(F, "  SSS_AddUVMapping(~s)\n", [UVArray]); false -> ok end,
  sss_subdivide(F, proplists:get_value(subdivisions,Attr)),
  Type = case proplists:get_value(export_normals,Attr) of
    true -> "Smooth"; false -> "Flat" end,
  io:format(F, "  SSS_BuildTextured~sMesh(~s)\n", [Type, TXArray]),
  io:put_chars(F,"}\n").




export_mesh2(F, Name, Vs, LnFs, Fs, ExportUVs, Ml, Mat, Ns, Tx, Attr)->
  db("Fs", Fs),
  Mesh2Label = "#declare " ++ Name ++ " = mesh2 {\n",
  io:put_chars(F, Mesh2Label),
%----------------- Vertex Vectors ---------------------------%
  io:format(F, "vertex_vectors {~p,", [length(Vs)]),
  comma_separated_list (F, fun({X,Y,Z}) -> io:format(F, "<~p,~p,~p>", [X,Y,Z]) end, Vs),
  io:put_chars(F, "}\n"),
%----------------- Normal Vectors ---------------------------%
  case proplists:get_value(export_normals, Attr) of
    true ->
      io:format(F, "normal_vectors {~p,", [length(Ns)]),
      comma_separated_list (F, fun({X,Y,Z}) -> io:format(F, "<~p,~p,~p>", [X,Y,Z]) end, Ns),
      io:put_chars(F, "}\n");
    false -> ok end,
%------------------ UV Vectors -------------------------------%
  case ExportUVs of
    true ->
      io:format(F, "uv_vectors { ~p,", [length(Tx)]),
      comma_separated_list (F, fun({U,V}) -> io:format(F, "<~p,~p>", [U,V]) end, Tx),
      io:put_chars(F, "}\n");
    false -> ok end,
%----------------- Texture List -------------------------------%
  io:format(F,"texture_list { ~p,\n", [length(Ml)]),
  foreach(fun({M0,Index}) ->
    SafeName = kayos_utils:safe_name(atom_to_list(M0)),
    TmpStr0 = "	texture { " ++ SafeName ++ "_WingsMat }\n",
    io:put_chars(F, TmpStr0)
  end, Ml),
  io:put_chars(F, "}\n"),
%--------------------- Face Indicies --------------------------%
  io:format(F, "face_indices {~p,", [LnFs]),
  comma_separated_list(F, fun ({_Ms,Faces}) ->
    comma_separated_list(F, fun(#e3d_face{vs=Vs0,mat=Mat0}) ->
      io:format(F, "<~p,~p,~p>", Vs0),
      MatIndex = material_index(Mat0, Ml),
      io:format(F, ",~p", [MatIndex]) end,
    Faces) end,
  Fs),
  io:put_chars(F, "}\n"),
%---------------------- Normal Indicies ------------------------------%
  case proplists:get_value(export_normals, Attr) of
    true ->
      io:format(F, "normal_indices {~p,", [LnFs]),
      comma_separated_list (F, fun({_Ms,Faces}) ->
        comma_separated_list (F, fun(#e3d_face{ns=Ns0}) ->
          io:format(F, "<~p,~p,~p>", Ns0) end,
	Faces) end,
      Fs),
      io:put_chars(F, "}\n");
    false -> io:put_chars("no Normals\n") end,
%----------------------- UV Indicies ---------------------------------%
  case ExportUVs of
    true ->
      io:format(F, "uv_indices {~p,", [LnFs]),
      comma_separated_list (F, fun({_Ms,Faces}) ->
        comma_separated_list (F, fun(#e3d_face{tx=Tx0}) ->
          io:format(F, "<~p,~p,~p>", Tx0) end,
	Faces) end,
      Fs),
      io:put_chars(F, "}\n");
    false -> ok end,
  case interior_list(Mat) of
    {true, InteriorName, InteriorAttr} -> export_interior(F,InteriorName, InteriorAttr);
    {false, _, _ } -> ok end,
  io:put_chars(F, "}\n").


%%%
%%% Utility Functions.
%%%


interior_list([{Name,Mat}|T]) ->
%io:format("Mat: ~p\n",[Mat]),
Povray = proplists:get_value(povray, Mat, []),
%io:format("Mat: ~p\n",[Mat]),
case proplists:get_value(pov_interior,Povray,false) of
  true -> {true, kayos_utils:safe_name(atom_to_list(Name)), Povray};
  false -> interior_list(T) end;
interior_list([]) -> {false, none, none}.


material_list(Mat)-> material_list(Mat,[],0).
material_list([{H,_}|T],Acc0,Incr) ->
	Acc = append([{H,Incr}],Acc0),
	material_list(T,Acc,Incr+1);
material_list([],Acc,_)-> reverse(Acc).

material_index([Mat|_],Ml) -> material_index(Mat,Ml,0).
material_index(Mat, [{Ref,Index}|T], Incr) ->
Incr0 = case Mat of
	Ref ->
		%io:format("Matched ~p\n", [Index]),
		Index;
	_->
		%io:put_chars("Not Matched\n"),
		Incr
	end,
	material_index(Mat, T, Incr0);
material_index(_,[],Index)-> Index.

start_array(F, ArrayName, NumItems, ItemLength) ->
  StartArray0 = "  #local " ++ ArrayName ++ " = array[" ++ integer_to_list(NumItems) ++ "]",
  StartArray = case ItemLength of
    1 -> StartArray0 ++ "{\n    ";
    _ -> StartArray0 ++ "[" ++ integer_to_list(ItemLength) ++ "]{\n  " end,
  io:put_chars(F, StartArray).
end_array(F) ->
  io:put_chars(F,"\n  }\n").

sss_subdivide(F,0)->
  ok;
sss_subdivide(F,Subdivisions)->
  db("subdivisions",Subdivisions),
  io:put_chars(F, "SSS_Subdivide()\n"),
  sss_subdivide(F,Subdivisions-1).



